/* Copyright 2008-2010 Yann Guermeur                                       */

/* This program is free software; you can redistribute it and/or modify     */
/* it under the terms of the GNU General Public License as published by     */
/* the Free Software Foundation; either version 2 of the License, or        */
/* (at your option) any later version.                                      */

/* This program is distributed in the hope that it will be useful,          */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            */
/* GNU General Public License for more details.                             */

/* You should have received a copy of the GNU General Public License        */
/* along with this program; if not, write to the Free Software              */
/* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA*/

/*--------------------------------------------------------------------------*/
/*  Name           : eval_SVM.c                                             */
/*  Version        : 1.0                                                    */
/*  Creation       : 06/20/08                                               */
/*  Last update    : 03/18/10                                               */
/*  Subject        : Implementation of the M-SVM^2                          */
/*  Module         : M-SVM used in test                                     */
/*  Author         : Yann Guermeur Yann.Guermeur@loria.fr                   */
/*--------------------------------------------------------------------------*/


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <ctype.h>
#include "algebre.h"
#include "biblio.h"

#define maximum(a,b) ((a)>=(b)?(a):(b))

#define true 1
#define false 0
#define taille 81
#define very_small 1e-6
#define step 100

FILE *fs, *fc;

long Q=0, i, j, k, l, dim_input, nb_data_app=0, nb_data_test=0, y_i,
ind_max, *y_app, *y_test, nature_kernel=0, *nb_sup_vect, **mat_conf;

char fichier_app[taille], fichier_alpha[taille],
fichier_test[taille], commande[taille],
fichier_fichcom[taille], fichier_resultat[taille];

double **X_app, **X_test, **alpha, *b_SVM, C=0.0, twoCinv=0.0,
primal=0.0, dual=0.0, Q_form_dual=0.0, **gradient, max_alpha=0.0,
**H_alpha, **H_tilde_alpha, alpha_H_tilde_alpha=0.0, *partial_average_alpha,
norm, *output, sum, partial, **xi_opt, **xi_est,
L_form_dual=0.0, maximum, ratio=0.0, xi_est_M_xi_est=0.0,
sum_all_alpha=0.0, sum_all_xi_opt=0.0, kernel=0.0, alpha_H_alpha=0.0;

/* Functions included in this program */

void caract_db();
void read_data();
void read_alpha();
void check_feasible_sol();
void compute_gradient();
void estime_b();
void compute_xi_opt_est();
void check_xi_opt();
void eval_training();
void compute_obj_dual();
void compute_obj_primal();
double fonction(long categorie, double *vecteur);
void compute_outputs(double **X, long *y, long nb_data);
void Pause(char *message);

main(int argc, char *argv[])

{

strcpy(fichier_fichcom, argv[1]);

system("clear");
caract_db();
read_data();
read_alpha();
check_feasible_sol();
compute_gradient();
estime_b();
compute_xi_opt_est();
check_xi_opt();
compute_obj_dual();
compute_obj_primal();
eval_training();
compute_outputs(X_test, y_test, nb_data_test);

}

void Pause(char *message)

{

if(message!=NULL)
  printf("%s\n",message);

printf("\a""=> New Line to proceed...\n");
while(getchar()!='\n');

return;

}

void caract_db()

{

if((fs=fopen(fichier_fichcom, "r"))==NULL)
  {
  printf("\nFile of parameters %s: cannot be open...\n", fichier_fichcom);
  exit(0);
  }

fscanf(fs, "%d", &Q);

mat_conf = matrix_l(Q,Q);

fscanf(fs, "%d", &nature_kernel);
fscanf(fs, "%lf", &C);
twoCinv = 1.0 / (2.0 * C);

fscanf(fs, "%s", fichier_app);
printf("\nThe file of training data is: %s", fichier_app);
fscanf(fs, "%s", fichier_test);
printf("\nThe file of test data is: %s", fichier_test);

fscanf(fs, "%s", fichier_alpha);
fscanf(fs, "%s", fichier_resultat);

fclose(fs);

}

void read_data()

{

double value_y;

if((fs=fopen(fichier_app, "r"))==NULL)
  {
  printf("\nFile of data %s: cannot be open...\n", fichier_app);
  exit(0);
  }

fscanf(fs, "%d", &nb_data_app);
fscanf(fs, "%d", &dim_input);

X_app = matrix(nb_data_app, dim_input);
y_app = (long *) calloc(nb_data_app+1, sizeof(long));

for(i=1; i<=nb_data_app; i++)
  {
  for(j=1; j<=dim_input; j++)
    fscanf(fs, "%lf", &X_app[i][j]);
  fscanf(fs, "%lf", &value_y);
  y_app[i] = (long) value_y;
  }

fclose(fs);

if((fs=fopen(fichier_test, "r"))==NULL)
  {
  printf("\nFile of data %s: cannot be open...\n", fichier_test);
  exit(0);
  }

fscanf(fs, "%d", &nb_data_test);
fscanf(fs, "%d", &dim_input);

X_test = matrix(nb_data_test, dim_input);
y_test = (long *) calloc(nb_data_test+1, sizeof(long));

for(i=1; i<=nb_data_test; i++)
  {
  for(j=1; j<=dim_input; j++)
    fscanf(fs, "%lf", &X_test[i][j]);
  fscanf(fs, "%lf", &value_y);
  y_test[i] = (long) value_y;
  }

fclose(fs);

printf("\nThe files of training and test data have been read...\n");

/*
display_mat(X_test, nb_data_test, dim_input, 6, 4);
Pause("");
*/

}

void read_alpha()

{

if((fs=fopen(fichier_alpha, "r"))==NULL)
  {
  printf("\nFile of dual variables %s: cannot be open...\n", 
  fichier_alpha);
  exit(0);
  }

alpha = matrix(nb_data_app, Q);
partial_average_alpha = (double *) calloc(nb_data_app+1, sizeof(double));

for(i=1; i<=nb_data_app; i++)
  {
  partial_average_alpha[i] = 0.0;
  for(k=1; k<=Q; k++)
    {
    fscanf(fs, "%lf", &alpha[i][k]);
    sum_all_alpha += alpha[i][k];
    partial_average_alpha[i] += alpha[i][k];

    if(alpha[i][k] > max_alpha)
      max_alpha = alpha[i][k];
    }
  partial_average_alpha[i] /= Q;
  }

fclose(fs);

printf("\nThe maximum value among the dual variables is: %e\n", max_alpha);

/*
display_mat(alpha, nb_data_app, Q, 9, 4);
Pause("");
*/

}

void check_feasible_sol()

{

double *constraints;

for(i=1; i<=nb_data_app; i++)
  for(k=1; k<=Q; k++)
    if(((k == y_app[i]) && (alpha[i][k] != 0.0)) || (alpha[i][k] < 0.0))
      {
      printf("\nNo feasible solution: alpha[%d][%d] = %lf\n\n",
             i, k, alpha[i][k]);
      exit(0);
      }
     
constraints = (double *) calloc(Q+1, sizeof(double));

for(k=1; k<=Q; k++)
  constraints[k] = sum_all_alpha / Q;

for(i=1; i<=nb_data_app; i++)
  for(k=1; k<=Q; k++)
    constraints[k] -= alpha[i][k];  

printf("\nSatisfaction of the equality constraints:\n\n");

for(k=1; k<=Q; k++)
  printf("%s%e\n", constraints[k] < 0.0 ? " " : "  ", constraints[k]);

norm = 0.0;

for(k=1; k<=Q; k++)
  norm += constraints[k] * constraints[k];

norm = sqrt(norm);

if(norm >= very_small)
  {
  printf("\nLarge deviation of the equality constraints...\n");
  Pause("");
  }

sprintf(commande, "cp %s Feasible/.", fichier_alpha);
system(commande);

}

void compute_gradient()

{

gradient = matrix(nb_data_app, Q);
H_alpha  = matrix(nb_data_app, Q);
H_tilde_alpha = matrix(nb_data_app, Q);

printf("\nStart of the gradient computation...\n");

for(i=1; i<=nb_data_app; i++)
  {
  if((i%step) == 0)
    printf("\n%9d", i);
  y_i = y_app[i];

  for(k=1; k<=Q; k++)
    {
    H_alpha[i][k] = 0.0;
    H_tilde_alpha[i][k] = 0.0;
    if(k != y_i)
      {
      for(j=1; j<=nb_data_app; j++)
        {
        partial = alpha[j][k] - partial_average_alpha[j];

	if(partial != 0.0)
          {
          kernel = ker(nature_kernel, X_app[i], X_app[j], dim_input);
	  H_alpha[i][k] += partial * kernel;
	  if(i == j)
            kernel += twoCinv;
	  H_tilde_alpha[i][k] += partial * kernel;
          }

        }
      }
    }
  }

for(i=1; i<=nb_data_app; i++)
  {
  y_i = y_app[i];
  for(k=1; k<=Q; k++)
    if(k != y_i)
      gradient[i][k] = - H_tilde_alpha[i][k] + (1.0 / (Q - 1.0));
    else
      gradient[i][k] = 0.0;
  }

/*
printf("\nComponent of the gradient vector\n");
display_mat(gradient, nb_data_app, Q, 9, 6);
Pause("");
*/

printf("\n\nEnd of the gradient computation...\n");

}

double fonction(long categorie, double *vecteur)

{

long index;

double resultat = 0.0;

for(index=1; index<=nb_data_app; index++)
  {
  partial = partial_average_alpha[index] - alpha[index][categorie];

  if(partial != 0.0)
  resultat += partial * ker(nature_kernel, X_app[index], vecteur, dim_input);
  }

return resultat;

}

void eval_training()

{

output = (double *) calloc(Q+1, sizeof(double));

for(k=1; k<=Q; k++)
  for(l=1; l<=Q; l++)
    mat_conf[k][l] = 0;

printf("\n\n*** Computing the training performance...\n");

for(i=1; i<=nb_data_app; i++)
  {
  if((i%step) == 0)
    printf("\n%9d", i);
  y_i = y_app[i];
  output[y_i] = 0.0;

  for(k=1; k<=Q; k++)
    if(k != y_i)
      {
      output[k] = -H_alpha[i][k] + b_SVM[k];
      output[y_i] -= output[k];
      }

  ind_max = 1;
  maximum = output[1];

  for(k=2; k<=Q; k++)
    if(output[k] > maximum)
      {
      maximum = output[k];
      ind_max = k;
      }

  mat_conf[y_i][ind_max]++;
  }

printf("\n\n*** Training performance\n");
display_full_stats(Q, mat_conf);
Pause("");

}

void compute_xi_opt_est()

{

xi_opt = matrix(nb_data_app, Q);
xi_est = matrix(nb_data_app, Q);

for(i=1; i<=nb_data_app; i++)
  {
  y_i = y_app[i];
  xi_opt[i][y_i] = 0.0;
  xi_est[i][y_i] = 0.0;

  for(k=1; k<=Q; k++)
    if(k != y_i)
      {
      xi_opt[i][k] = twoCinv * (alpha[i][k] - partial_average_alpha[i]);
      sum_all_xi_opt += xi_opt[i][k];
      xi_est[i][k] = maximum(xi_opt[i][k], 
		     -H_alpha[i][k] + b_SVM[k] + (1.0 / (Q - 1.0)));
      }
  }

/*
printf("\nMatrix of the computed values of the slack variables\n");
display_mat(xi_opt, nb_data_app, Q, 9, 4);
Pause("");
*/

}

void check_xi_opt()

{

for(i=1; i<=nb_data_app; i++)
  {
  y_i = y_app[i];

  for(k=1; k<=Q; k++)
    if(k != y_i)
      {
      partial = H_alpha[i][k] - (1.0 / (Q - 1.0)) + xi_opt[i][k];
      partial += gradient[i][k];
      if(fabs(partial) > very_small)
        {
        printf("\nProblem in the computation of the slack variables...\n");
        exit(0);
        }
      }
  }

printf("\nThe values of the slack variables have been checked...\n");
/* Pause(""); */

}

void estime_b()

{

nb_sup_vect = (long *) calloc(Q+1, sizeof(long));
b_SVM = (double *) calloc(Q+1, sizeof(double));

for(k=1; k<=Q; k++)
  {
  nb_sup_vect[k] = 0;
  b_SVM[k] = 0.0;
  }

sum = 0.0;

printf("\nVector b before centering\n\n");

for(i=1; i<=nb_data_app; i++)
  for(k=1; k<=Q; k++)
    if(alpha[i][k] > 0)
      {
      nb_sup_vect[k]++;
      b_SVM[k] -= gradient[i][k];
      }

for(k=1; k<=Q; k++)
  {
  if(nb_sup_vect[k] > 0)
    b_SVM[k] /= (double) nb_sup_vect[k];
  printf(" b_%1d = %c%e\n", k, (b_SVM[k] >= 0) ? ' ': '-', fabs(b_SVM[k]));
  sum += b_SVM[k];
  }

printf("\nSum of the components of b: %e\n", sum);
Pause("");

if(sum != 0.0)
  {
  sum /= Q;
  for(k=1; k<=Q; k++)
    b_SVM[k] -= sum;
  }

}

void compute_outputs(double **X, long *y, long nb_data)

{

printf("\n*** Computing the test performance...\n");

for(k=1; k<=Q; k++)
  for(l=1; l<=Q; l++)
    mat_conf[k][l] = 0;

if((fc=fopen(fichier_resultat, "w"))==NULL)
  {
  printf("\nFile of outputs %s: cannot be open...\n", fichier_resultat);
  exit(0);
  }

for(i=1; i<=nb_data; i++)
  {
  for(k=1; k<=Q; k++)
    output[k] = fonction(k, X[i]) + b_SVM[k];

  ind_max = 1;
  maximum = output[1];

  for(k=2; k<=Q; k++)
    if(output[k] > maximum)
      {
      maximum = output[k];
      ind_max = k;
      }

  mat_conf[y[i]][ind_max]++;

  for(k=1; k<=Q; k++) 
    fprintf(fc, "%12.6f%c", output[k], (k==Q) ? '\n' : ' ');

  if((i % step) == 0)
    {
    printf("\nExample: %5d", i);
    display_stats(Q, mat_conf);
    }
  }

fclose(fc);

printf("\n*** Test performance\n");
display_full_stats(Q, mat_conf);

}

void compute_obj_dual()

{

for(i=1; i<=nb_data_app; i++)
  for(k=1; k<=Q; k++)
    {
    alpha_H_alpha += H_alpha[i][k] * alpha[i][k];
    alpha_H_tilde_alpha += H_tilde_alpha[i][k] * alpha[i][k];
    }

Q_form_dual = alpha_H_tilde_alpha;
L_form_dual = sum_all_alpha / (Q - 1.0);

/*
if(Q_form_dual > 0.0)
  {
  printf("\nComponents of the dual objective function: (Q)%lf - (L)%lf",
         Q_form_dual, L_form_dual);
  ratio = L_form_dual / Q_form_dual;
  printf(
  "\nRatio of the components of the dual objective function (-> 1.0): %lf\n", 
  ratio);
  }
*/

dual = -0.5 * Q_form_dual + L_form_dual;

}

void compute_obj_primal()

{

for(i=1; i<=nb_data_app; i++)
  for(k=1; k<=Q; k++)
    for(l=1; l<=Q; l++)
      {
      partial = (k == l) ? 2.0 : 1.0;
      xi_est_M_xi_est += partial * xi_est[i][k] * xi_est[i][l];
      }

/*
printf("\nEmpirical contribution to the primal objective function: %lf", 
       C * xi_M_xi);
*/

if(fabs(sum_all_alpha - 2.0 * C * Q * sum_all_xi_opt) < very_small)
  printf("1^T alpha = 2CQ 1^T xi = %e\n", sum_all_alpha);

primal = 0.5 * alpha_H_alpha + C * xi_est_M_xi_est;

printf("\n            Dual objective function: %e", dual);
printf("\nEstimated primal objective function: %e", primal);

if(primal != 0.0)
  printf("\n(ratio %5.2lf\%)\n", 100.0 * dual / primal);

Pause("");

}
