/* Copyright 2008, 2009 Yann Guermeur                                       */

/* This program is free software; you can redistribute it and/or modify     */
/* it under the terms of the GNU General Public License as published by     */
/* the Free Software Foundation; either version 2 of the License, or        */
/* (at your option) any later version.                                      */

/* This program is distributed in the hope that it will be useful,          */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            */
/* GNU General Public License for more details.                             */

/* You should have received a copy of the GNU General Public License        */
/* along with this program; if not, write to the Free Software              */
/* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA*/

/*--------------------------------------------------------------------------*/
/*  Name           : train_SVM.c                                            */
/*  Version        : 1.0                                                    */
/*  Creation       : 06/20/08                                               */
/*  Last update    : 11/20/09                                               */
/*  Subject        : Training algorithm of the M-SVM^2                      */
/*  Algo.          : Frank-Wolfe algorithm + decomposition method           */
/*  Author         : Yann Guermeur Yann.Guermeur@loria.fr                   */
/*--------------------------------------------------------------------------*/


#include <stdio.h>
#include <stdlib.h> 
#include <string.h>
#include <math.h>
#include <ctype.h>
#include "algebre.h"

#define minimum(a,b) ((a)<=(b)?(a):(b))
#define maximum(a,b) ((a)>=(b)?(a):(b))

#define true 1
#define false 0
#define taille 81
#define very_small 1e-3
#define very_large 1e3
#define nb_iter 1e9
#define step 100

FILE *fs, *fc;

unsigned short xi[3];

long jump=false, nature_kernel=0,
i, j, k, l, m, dim_input, nb_data, iter, *y, row,
y_i, y_j, Q=0, chunk_size=0, *table_chunk, *out_of_chunk, *in_chunk,
ind_pattern, nrand48(), random_num, rank, nb_points_chunk, SV, nb_SV;

char fichier_data[taille], fichier_alpha[taille], letter, commande[taille],
fichier_lp_sol[taille]="lp.res",
*ligne, *frag, fichier_fichcom[taille], fichier_alpha_0[taille],
*tampon[6] = {"00000","0000","000","00","0",""},
fichier_lp[taille]="lp.lp";

double **X, **alpha, C=0.0, twoCinv=0.0, **gradient, **delta, value,
obj_lp_lu=0.0, **lp_sol, theta_opt=0.0,
partiel, **K, **H_delta, gradient_alpha, gradient_gamma, *rhs,
gradient_diff;

/* Functions included in this program */

long indice_tampon(long parametre);
void caract_db();
void alloc_memory();
void read_data();
void read_alpha();
void compute_table_chunk();
void compute_K();
void compute_gradient();
void write_lp_file();
void read_lp_sol();
void check_opt_sol();
void compute_delta();
void compute_H_delta();
void compute_theta_opt();
void compute_new_alpha();
void write_vector(double **vector, char *fichier);
void Pause(char *message);
void compute_objective_function();

main(int argc, char *argv[])

{

system("clear");
srand48(getpid());
strcpy(fichier_fichcom, argv[1]);

caract_db();
read_data();
read_alpha();

for(iter=1; iter<=nb_iter; iter++)
  {

  if((iter%step) == 0)
    printf("\n\n*** Iteration: %6d\n", iter);
  jump = false;
  compute_table_chunk();
  compute_K();
  compute_gradient();
  if(jump == false)
    {
    write_lp_file();
    system("./lp_solve < lp.lp > lp.res 2> lp.err");
    read_lp_sol();
    }
  if(jump == false)
    check_opt_sol();
  if(jump == false)
    {
    compute_delta();
    compute_theta_opt();
    compute_new_alpha();
    }
  if((iter%step) == 0)
    {
    write_vector(alpha, fichier_alpha);
/*
    compute_objective_function();
    Pause(""); 
*/
    }

  } 

}

long indice_tampon(long parametre)

{

long valeur_long;

valeur_long = (long) floor(log10((double) parametre));
return valeur_long;

}

void Pause(char *message)

{

char c;
static int n = 0;

if(message!=NULL)
  printf("%s\n",message);

printf("\a\n=> New Line to proceed...\n");

n = read(0, &c, 1);

while(c != '\n')
  n = read(0, &c, 1);

return;

}

void caract_db()

{

if((fs=fopen(fichier_fichcom, "r"))==NULL)
  {
  printf("\nFile of parameters %s: cannot be open...\n", fichier_fichcom);
  exit(0);
  }

fscanf(fs, "%d", &Q);
fscanf(fs, "%d", &nature_kernel);
fscanf(fs, "%lf", &C);
twoCinv = 1.0 / (2.0 * C);
fscanf(fs, "%d", &chunk_size);
fscanf(fs, "%s", fichier_data);
printf("\nThe data file is: %s\n", fichier_data);
fscanf(fs, "%s", fichier_alpha_0);
fscanf(fs, "%s", fichier_alpha);

fclose(fs);

sprintf(commande, "cp %s Save_alpha/.", fichier_alpha);

}

void alloc_memory()

{

X = matrix(nb_data, dim_input);
y = (long *) calloc(nb_data+1, sizeof(long));
alpha = matrix(nb_data, Q);
out_of_chunk = (long *) calloc(nb_data+1, sizeof(long));
in_chunk = (long *) calloc(nb_data+1, sizeof(long));
gradient = matrix(chunk_size, Q);
lp_sol = matrix(chunk_size, Q);
delta = matrix(chunk_size, Q);
table_chunk = (long *) calloc(chunk_size+1, sizeof(long));
K = matrix(chunk_size, nb_data);
H_delta = matrix(chunk_size, Q);
rhs = (double *) calloc(Q+1, sizeof(double));
ligne = (char *) calloc(taille+1, sizeof(char));
frag = (char *) calloc(taille+1, sizeof(char));

}

void read_data()

{

double value_y;

if((fs=fopen(fichier_data, "r"))==NULL)
  {
  printf("\nData file %s: cannot be open...\n", fichier_data);
  exit(0);
  }

fscanf(fs, "%d", &nb_data);
fscanf(fs, "%d", &dim_input);

alloc_memory();

for(i=1; i<=nb_data; i++)
  {
  for(j=1; j<=dim_input; j++)
    fscanf(fs, "%lf", &X[i][j]);
  fscanf(fs, "%lf", &value_y);
  y[i] = (long) value_y;
  }

fclose(fs);

printf("\nThe file of the data has been read...\n");

/*
display_mat(X, nb_data, dim_input, 6, 4);
Pause("");
*/

}

void read_alpha()

{

if((fs=fopen(fichier_alpha_0, "r"))==NULL)
  {
  printf(
  "\nFile of the initial values of the dual variables %s: cannot be open...\n",
  fichier_alpha_0);
  exit(0);
  }

for(i=1; i<=nb_data; i++)
  for(k=1; k<=Q; k++)
    fscanf(fs, "%lf", &alpha[i][k]);

fclose(fs);

printf(
"\nThe file of the initial values of the dual variables has been read...\n");

}

void write_lp_file()

{

if((fc=fopen(fichier_lp, "w"))==NULL)
  {
  printf("\nFile lp %s: cannot be open...\n", fichier_lp);
  exit(0);
  }

fprintf(fc, "min:");

for(i=1; i<=chunk_size; i++)
  {
  ind_pattern = table_chunk[i];
  y_i = y[ind_pattern];
  for(k=1; k<=Q; k++)
    if(k != y_i)
      {
      if(gradient[i][k] > 0.0)
        fprintf(fc, " +%lf A0%s%d", gradient[i][k],
        tampon[indice_tampon(Q*(i-1)+k)], (Q*(i-1)+k));
      if(gradient[i][k] < 0.0)
        fprintf(fc, " %lf A0%s%d", gradient[i][k],
        tampon[indice_tampon(Q*(i-1)+k)], (Q*(i-1)+k));
      }
  }

fprintf(fc, ";\n");

for(k=1; k<Q; k++)
  {
  rhs[k] = 0.0;
  for(i=1; i<=nb_data; i++)
    if(in_chunk[i] == 0)
      {
      for(l=1; l<=Q; l++)
        rhs[k] += alpha[i][l];
      rhs[k] -= Q * alpha[i][k];
      }
  }

for(k=1; k<Q; k++)
  {
  fprintf(fc, "CONST%1d: ", k);

  for(i=1; i<=chunk_size; i++)
    {
    ind_pattern = table_chunk[i];
    y_i = y[ind_pattern];
    for(l=1; l<=Q; l++)
      if(l != y_i)
        {
        partiel = -1.0;
	if(l == k)
          partiel += Q;
        if(l != k)
          {
          fprintf(fc, "%f A0%s%d ", partiel,
          tampon[indice_tampon(Q*(i-1)+l)],(Q*(i-1)+l));
          }
        else
          fprintf(fc, "+%f A0%s%d ", partiel,
          tampon[indice_tampon(Q*(i-1)+l)], (Q*(i-1)+l));
	}
    }

  fprintf(fc, "= %lf;\n", rhs[k]);
  }

for(i=1; i<=chunk_size; i++)
  {
  ind_pattern = table_chunk[i];
  y_i = y[ind_pattern];
  for(k=1; k<=Q; k++)
    if(k != y_i)
      fprintf(fc, "A0%s%d < %lf;\n",
      tampon[indice_tampon(Q*(i-1)+k)], (Q*(i-1)+k), very_large);
  }

fclose(fc);

}

void compute_table_chunk()

{

for(i=1; i<=nb_data; i++)
  {
  out_of_chunk[i] = i;
  in_chunk[i] = 0;
  }

nb_SV = 0;

for(i=1; i<=nb_data; i++)
  {
  SV = 0;
  k = 1;
  while((SV == 0) && (k <= Q))
    {
    if(alpha[i][k] > 0.0)
      {
      SV = 1;
      nb_SV++;
      }
    k++;
    }
  }

if((iter%step) == 0)
  printf("\nNumber of support vectors: %d", nb_SV);

for(i=1; i<=chunk_size; i++)
  {
  random_num = nrand48(xi);
  row = (random_num % (nb_data-i+1))+1;
  table_chunk[i] = out_of_chunk[row];
  for(j=row; j<=nb_data-i; j++)
    out_of_chunk[j] = out_of_chunk[j+1];
  in_chunk[table_chunk[i]] = 1;
  }

}

void compute_K()

{

for(i=1; i<=chunk_size; i++)
  {
  ind_pattern = table_chunk[i];
  for(j=1; j<=nb_data; j++)
    K[i][j] = ker(nature_kernel, X[ind_pattern], X[j], dim_input);
  K[i][ind_pattern] += twoCinv;
  }

}

void compute_gradient()

{

double partial_sum_alpha = 0.0;

for(i=1; i<=chunk_size; i++)
  for(k=1; k<=Q; k++)
    {
    if(k != y[table_chunk[i]])
      {
      gradient[i][k] = -1.0 / (Q - 1.0);
      for(j=1; j<=nb_data; j++)
        {
	partial_sum_alpha = 0.0;

        for(l=1; l<=Q; l++)
          partial_sum_alpha += alpha[j][l];

	partiel = alpha[j][k] - (partial_sum_alpha / Q);
	gradient[i][k] += partiel * K[i][j];
        }
      }
    }

}

void compute_H_delta()

{

double partial_sum_delta = 0.0;

for(i=1; i<=chunk_size; i++)
  for(k=1; k<=Q; k++)
    {
    H_delta[i][k] = 0.0; 
    if(k != y[table_chunk[i]])
      {
      for(j=1; j<=chunk_size; j++)
        {
        partial_sum_delta = 0.0;

        for(l=1; l<=Q; l++)
          partial_sum_delta += delta[j][l];

        partiel = delta[j][k] - (partial_sum_delta / Q);
	H_delta[i][k] += partiel * K[i][table_chunk[j]];
	}
      }
    }

}

void read_lp_sol()

{

if((fs=fopen(fichier_lp_sol, "r"))==NULL)
  {
  printf("\nThe file %s of the LP solution cannot be open...\n",
  fichier_lp_sol);
  exit(0);
  }

fgets(ligne, taille, fs);

if(strncmp("This problem is infeasible", ligne, 26) == 0)
  {
  printf("\nProblem with the LP...\n");
  jump=true;
  }
else
  {
  frag = strtok(ligne, ":");
  frag = strtok(NULL, ":");
  obj_lp_lu = atof(frag);
/*
printf("\nObjective function of the LP problem (value read): %lf", obj_lp_lu);
*/

  for(i=1; i<=chunk_size; i++)
    {
    ind_pattern = table_chunk[i];
    y_i = y[ind_pattern];
    lp_sol[i][y_i]=0.0;
    }

  for(i=1; i<=chunk_size; i++)
    for(k=1; k<Q; k++)
      {
      fscanf(fs, "%c", &letter);
      fscanf(fs, "%d", &rank);
      fscanf(fs, "%lf", &value);
      j = (rank-1)/Q+1;
      l = (rank-1)%Q+1;
      lp_sol[j][l] = value;
/*
       printf("\nLetter: %c, rank: %d, value: %lf\n",
       letter, rank, value);
       printf("\n %d -> [%d][%d]\n", rank, j, l);
*/
      fscanf(fs, "%c", &letter);
      }
/*  Pause(""); */
  }

fclose(fs);

/*
printf("\nSolution of the LP problem:\n");

display_mat(lp_sol, chunk_size, Q, 14, 6);
Pause("");
*/

}

void check_opt_sol()

{

gradient_alpha = 0.0; 
gradient_gamma = 0.0;
gradient_diff = 0.0;

for(i=1; i<=chunk_size; i++)
  {
  ind_pattern = table_chunk[i];
  y_i = y[ind_pattern];

  for(k=1; k<=Q; k++)
    if(k != y_i)
      {
      gradient_alpha += gradient[i][k] * alpha[ind_pattern][k];
      gradient_gamma += gradient[i][k] * lp_sol[i][k];
      }
  }

gradient_diff = gradient_alpha - gradient_gamma;

if(gradient_diff <= 0.0)
  {
  if(gradient_diff < 0.0)
    {
    printf("\n\nInappropriate subset selection:");
    printf("\ngradient^T (alpha - gamma): %e", gradient_diff);
    }
  jump = true;
  }

/*
if(fabs((gradient_gamma - obj_lp_lu) / obj_lp_lu) > very_small)
  {
  printf("\nDeviation in the values of the objective function,");
  printf(" gradient^T gamma = %lf versus %lf (read)\n", 
  gradient_gamma, obj_lp_lu);
  }
*/

}

void compute_new_alpha()

{

for(i=1; i<=chunk_size; i++)
  {
  ind_pattern = table_chunk[i];
  for(k=1; k<=Q; k++)
    {
    alpha[ind_pattern][k] *= (1.0-theta_opt);
    alpha[ind_pattern][k] += theta_opt * lp_sol[i][k];
    }
  }

}

void compute_delta()

{

for(i=1; i<=chunk_size; i++)
  {
  ind_pattern = table_chunk[i];
  for(k=1; k<=Q; k++)
    delta[i][k] = alpha[ind_pattern][k] - lp_sol[i][k];
  }

}

void compute_theta_opt()

{

double denominateur = 0.0;

compute_H_delta();

for(i=1; i<=chunk_size; i++)
  for(k=1; k<=Q; k++)
    denominateur += delta[i][k] * H_delta[i][k];

if(denominateur <= 0.0)
  {
  printf("\nWrong estimate of the quadratic form...\n");
  exit(0);
  }

theta_opt = gradient_diff / denominateur;

if(theta_opt > 1.0)
  theta_opt = 1.0;

/*
if((iter%step) == 0)
  printf("\n    Optimal value of theta: %10.8f\n", theta_opt);
*/

}

void write_vector(double **vector, char *fichier)

{

if((fc=fopen(fichier, "w"))==NULL)
  {
  printf("\nThe file %s cannot be open...\n", fichier);
  exit(0);
  }

for(i=1; i<=nb_data; i++)
  for(k=1; k<=Q; k++)
    fprintf(fc, "%15.10f %c", vector[i][k], (k==Q) ? '\n': ' ');

fclose(fc);

system(commande);

}

void compute_objective_function()

{

double Q_term=0.0, L_term=0.0, current_value=0.0, 
       partial_1, partial_2, kernel;

for(i=1; i<=nb_data; i++)
  for(j=1; j<=nb_data; j++)
    {
    partial_1 = 0.0;
    for(k=1; k<=Q; k++)
      for(l=1; l<=Q; l++)
        {
        partial_2 = -1.0;
        if(k == l) 
          partial_2 += Q;
	partial_1 += partial_2 * alpha[i][k] * alpha[j][l];
        }
    if(partial_1 != 0.0)
      {
      kernel = ker(nature_kernel, X[i], X[j], dim_input); 
      if(i == j)
        kernel += twoCinv;
      Q_term += partial_1 * kernel;
      }
    }

Q_term /= Q;

for(i=1; i<=nb_data; i++)
  for(k=1; k<=Q; k++)
    L_term += alpha[i][k];

L_term /= (Q - 1.0);

current_value = -0.5 * Q_term + L_term;
printf("\n           Quadratic term: %lf", Q_term);
printf("\n              Linear term: %lf", L_term);
printf("\n       Objective function: %lf", current_value);

}
